# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests tree_rich_types.py module."""

from absl.testing import absltest
from absl.testing import parameterized
from orbax.checkpoint._src.metadata import tree_rich_types
from orbax.checkpoint._src.testing import test_tree_utils


class TreeRichTypesTest(parameterized.TestCase):

  @parameterized.product(
      bad_tree=[
          {"a": test_tree_utils.MyClass()},
      ]
  )
  def test_invalid_pytree_leaf(self, bad_tree):
    with self.assertRaisesRegex(
        ValueError, "Expected ValueMetadataEntry, got metadata pytree leaf"
    ):
      tree_rich_types.value_metadata_tree_to_json_str(bad_tree)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for CheckpointerHandler type registry."""

from absl.testing import absltest
from absl.testing import parameterized
from etils import epath
from orbax.checkpoint._src.handlers import checkpoint_handler
from orbax.checkpoint._src.handlers import handler_type_registry
from orbax.checkpoint._src.handlers import standard_checkpoint_handler


HandlerTypeRegistry = handler_type_registry.HandlerTypeRegistry


class TestHandler(checkpoint_handler.CheckpointHandler):
  def save(self, directory: epath.Path, *args, **kwargs):
    pass

  def restore(self, directory: epath.Path, *args, **kwargs):
    pass


class ParentHandler(checkpoint_handler.CheckpointHandler):
  class TestHandler(checkpoint_handler.CheckpointHandler):
    def save(self, directory: epath.Path, *args, **kwargs):
      pass

    def restore(self, directory: epath.Path, *args, **kwargs):
      pass


class StandardCheckpointHandler(checkpoint_handler.CheckpointHandler):
  def save(self, directory: epath.Path, *args, **kwargs):
    pass

  def restore(self, directory: epath.Path, *args, **kwargs):
    pass


class ChildStandardCheckpointHandler(
    standard_checkpoint_handler.StandardCheckpointHandler
):
  pass


class TypestrOverrideHandler(checkpoint_handler.CheckpointHandler):
  @classmethod
  def typestr(cls) -> str:
    return 'typestr_override'


class HandlerTypeRegistryTest(parameterized.TestCase):

  def test_register_and_get(self):
    registry = HandlerTypeRegistry()
    registry.add(TestHandler.typestr(), TestHandler)
    self.assertEqual(
        registry.get(TestHandler.typestr()),
        TestHandler,
    )
    registry.add(ParentHandler.TestHandler.typestr(), ParentHandler.TestHandler)
    self.assertEqual(
        registry.get(ParentHandler.TestHandler.typestr()),
        ParentHandler.TestHandler,
    )
    self.assertTrue(
        '__main__.TestHandler' in registry._registry
        or
        'handler_type_registry_test.TestHandler' in registry._registry
    )
    self.assertTrue(
        '__main__.ParentHandler.TestHandler' in registry._registry
        or
        'handler_type_registry_test.ParentHandler.TestHandler'
        in registry._registry
    )

  def test_register_different_modules(self):
    registry = HandlerTypeRegistry()
    registry.add(StandardCheckpointHandler.typestr(), StandardCheckpointHandler)
    self.assertEqual(
        registry.get(StandardCheckpointHandler.typestr()),
        StandardCheckpointHandler,
    )
    registry.add(
        standard_checkpoint_handler.StandardCheckpointHandler.typestr(),
        standard_checkpoint_handler.StandardCheckpointHandler
    )
    self.assertEqual(
        registry.get(
            standard_checkpoint_handler.StandardCheckpointHandler().typestr()
        ),
        standard_checkpoint_handler.StandardCheckpointHandler,
    )
    self.assertTrue(
        '__main__.StandardCheckpointHandler' in registry._registry
        or
        'handler_type_registry_test.StandardCheckpointHandler'
        in registry._registry
    )
    self.assertIn(
        'orbax.checkpoint._src.handlers.standard_checkpoint_handler.'
        'StandardCheckpointHandler',
        registry._registry
    )

  def test_register_duplicate_handler_type(self):
    registry = HandlerTypeRegistry()
    registry.add(TestHandler.typestr(), TestHandler)

    # Register the same handler and typestr again. OK.
    registry.add(TestHandler.typestr(), TestHandler)

    # Register a different handler with the same typestr. Error.
    with self.assertRaisesRegex(
        ValueError,
        'Handler type string '
        r'"(?:__main__|handler_type_registry_test)\.TestHandler"'
        ' already exists in the registry with type '
        r'<class \'(?:__main__|handler_type_registry_test)\.TestHandler\'>. '
        'Cannot add type '
        r'<class \'(?:__main__|handler_type_registry_test)\.'
        'ParentHandler.TestHandler\'>.',
    ):
      registry.add(TestHandler.typestr(), ParentHandler.TestHandler)

  def test_get_handler_type_not_found(self):
    registry = HandlerTypeRegistry()
    with self.assertRaisesRegex(
        ValueError,
        'Handler type string '
        r'"(?:__main__|handler_type_registry_test)\.TestHandler"'
        ' not found in the registry.',
    ):
      registry.get(TestHandler.typestr())

  def test_register_subclass_handler_type(self):
    registry = HandlerTypeRegistry()
    registry.add(
        standard_checkpoint_handler.StandardCheckpointHandler.typestr(),
        standard_checkpoint_handler.StandardCheckpointHandler
    )
    registry.add(
        ChildStandardCheckpointHandler.typestr(),
        ChildStandardCheckpointHandler
    )
    self.assertEqual(
        registry.get(
            standard_checkpoint_handler.StandardCheckpointHandler.typestr()
        ),
        standard_checkpoint_handler.StandardCheckpointHandler,
    )
    self.assertEqual(
        registry.get(ChildStandardCheckpointHandler.typestr()),
        ChildStandardCheckpointHandler,
    )

  def test_typestr_override(self):
    registry = HandlerTypeRegistry()
    registry.add(TypestrOverrideHandler.typestr(), TypestrOverrideHandler)
    self.assertEqual(
        registry.get(TypestrOverrideHandler.typestr()),
        TypestrOverrideHandler,
    )

if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized
import jax
import numpy as np
from orbax.checkpoint._src.arrays import abstract_arrays
from orbax.checkpoint._src.serialization import type_handlers

to_shape_dtype_struct = abstract_arrays.to_shape_dtype_struct


class AbstractArraysTest(parameterized.TestCase):

  @parameterized.parameters((None,), (np.float64,))
  def test_jax_array(self, target_dtype):
    target_dtype = target_dtype or np.float32
    arr = jax.device_put(
        np.arange(4, dtype=np.float32),
        jax.sharding.SingleDeviceSharding(jax.devices()[0]),
    )
    self.assertIsInstance(arr, jax.Array)
    sds = to_shape_dtype_struct(arr, dtype=target_dtype)
    self.assertIsInstance(sds, jax.ShapeDtypeStruct)
    self.assertEqual(sds.shape, (4,))
    self.assertEqual(sds.dtype, target_dtype)
    self.assertEqual(
        sds.sharding, jax.sharding.SingleDeviceSharding(jax.devices()[0])
    )

  @parameterized.parameters((None,), (np.float64,))
  def test_shape_dtype_struct(self, target_dtype):
    target_dtype = target_dtype or np.float32
    arr = jax.ShapeDtypeStruct(
        (4,),
        np.float32,
        sharding=jax.sharding.SingleDeviceSharding(jax.devices()[0]),
    )
    sds = to_shape_dtype_struct(arr, dtype=target_dtype)
    self.assertIsInstance(sds, jax.ShapeDtypeStruct)
    self.assertEqual(sds.shape, (4,))
    self.assertEqual(sds.dtype, target_dtype)
    self.assertEqual(
        sds.sharding, jax.sharding.SingleDeviceSharding(jax.devices()[0])
    )

  @parameterized.parameters((None,), (np.float64,))
  def test_restore_args(self, target_dtype):
    target_dtype = target_dtype or np.float32
    args = type_handlers.ArrayRestoreArgs(
        global_shape=(4,),
        dtype=np.float32,
        sharding=jax.sharding.SingleDeviceSharding(jax.devices()[0]),
    )
    sds = to_shape_dtype_struct(args, dtype=target_dtype)
    self.assertIsInstance(sds, jax.ShapeDtypeStruct)
    self.assertEqual(sds.shape, (4,))
    self.assertEqual(sds.dtype, target_dtype)
    self.assertEqual(
        sds.sharding, jax.sharding.SingleDeviceSharding(jax.devices()[0])
    )

  @parameterized.parameters((None,), (np.float64,))
  def test_numpy_array(self, target_dtype):
    target_dtype = target_dtype or np.float32
    arr = np.arange(4, dtype=np.float32)
    sds = to_shape_dtype_struct(arr, dtype=target_dtype)
    self.assertIsInstance(sds, jax.ShapeDtypeStruct)
    self.assertEqual(sds.shape, (4,))
    self.assertEqual(sds.dtype, target_dtype)
    self.assertIsNone(sds.sharding, None)

  @parameterized.product(
      value=(5, np.int32(5)),
      scalar_dtype=(None, int, np.int32, np.int64, float, np.float32),
  )
  def test_int(self, value, scalar_dtype):
    sds = to_shape_dtype_struct(value, scalar_dtype=scalar_dtype)
    target_dtype = scalar_dtype or type(value)
    self.assertEqual(type(sds), target_dtype)
    self.assertEqual(sds, target_dtype(value))

  @parameterized.product(
      value=(5.5, np.float32(5.5)),
      scalar_dtype=(None, float, np.float32, np.float64, int, np.int32),
  )
  def test_float(self, value, scalar_dtype):
    sds = to_shape_dtype_struct(value, scalar_dtype=scalar_dtype)
    target_dtype = scalar_dtype or type(value)
    self.assertEqual(type(sds), target_dtype)
    self.assertEqual(sds, target_dtype(value))


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for serializing and deserializing metadata."""

from typing import Any, Sequence


def validate_type(obj: Any, field_type: type[Any]):
  if not isinstance(obj, field_type):
    raise ValueError(f'Object must be of type {field_type}, got {type(obj)}.')


def validate_field(
    obj: Any,
    field_name: str,
    field_type: type[Any] | Sequence[type[Any]]
):
  """Validates a single field in a dictionary.

  field_type can optionally be a sequence of types, in which case the field
  must be of any one of the types in the sequence.

  Args:
    obj: The object to validate.
    field_name: The name of the field to validate.
    field_type: The type (or sequence of types) of the field to validate.
  """
  if field_name not in obj or obj[field_name] is None:
    return
  field = obj[field_name]
  if isinstance(field_type, Sequence):
    if not any(isinstance(field, f_type) for f_type in field_type):
      raise ValueError(
          f'Metadata field "{field_name}" must be any one of '
          f'types {list(field_type)}, got {type(field)}.'
      )
  elif not isinstance(field, field_type):
    raise ValueError(
        f'Metadata field "{field_name}" must be of type {field_type}, '
        f'got {type(field)}.'
    )


def validate_dict_entry(
    dict_field: dict[Any, Any],
    dict_field_name: str,
    key: Any,
    key_type: type[Any],
    value_type: type[Any] | None = None,
):
  """Validates a single entry in a dictionary field."""
  if not isinstance(key, key_type):
    raise ValueError(
        f'Metadata field "{dict_field_name}" keys must be of type {key_type}, '
        f'got {type(key)}.'
    )
  if value_type is not None:
    dict_field = dict_field[dict_field_name]
    if not isinstance(dict_field[key], value_type):
      raise ValueError(
          f'Metadata field "{dict_field_name}" values must be of '
          f'type {value_type}, got {type(dict_field[key])}.'
      )



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import pickle
from typing import Any, Callable

from absl.testing import absltest
from absl.testing import parameterized
from etils import epath
from orbax.checkpoint._src.logging import step_statistics
from orbax.checkpoint._src.metadata import checkpoint
from orbax.checkpoint._src.metadata import root_metadata_serialization
from orbax.checkpoint._src.metadata import step_metadata_serialization


StepMetadata = checkpoint.StepMetadata
RootMetadata = checkpoint.RootMetadata
MetadataStore = checkpoint.MetadataStore
StepStatistics = step_statistics.SaveStepStatistics
CompositeItemMetadata = checkpoint.CompositeItemMetadata
SingleItemMetadata = checkpoint.SingleItemMetadata

_SAMPLE_FORMAT = 'sample_format'


class CheckpointMetadataTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.directory = epath.Path(self.create_tempdir().full_path)
    self._non_blocking_store_for_writes = checkpoint.metadata_store(
        enable_write=True, blocking_write=False
    )

  def tearDown(self):
    super().tearDown()
    checkpoint._METADATA_STORE_FOR_WRITES.close()
    checkpoint._METADATA_STORE_FOR_READS.close()
    self._non_blocking_store_for_writes.close()
    checkpoint._METADATA_STORE_NON_BLOCKING_FOR_READS.close()

  def write_metadata_store(
      self, blocking_write: bool
  ) -> MetadataStore:
    if blocking_write:
      return checkpoint.metadata_store(
          enable_write=True, blocking_write=True
      )
    return self._non_blocking_store_for_writes

  def read_metadata_store(
      self, blocking_write: bool
  ) -> MetadataStore:
    return checkpoint.metadata_store(
        enable_write=False, blocking_write=blocking_write
    )

  def deserialize_metadata(
      self,
      metadata_type: type[StepMetadata] | type[RootMetadata],
      metadata: checkpoint.SerializedMetadata,
      item_metadata: CompositeItemMetadata | SingleItemMetadata | None = None,
      metrics: dict[str, Any] | None = None,
  ) -> StepMetadata | RootMetadata:
    if metadata_type == StepMetadata:
      return step_metadata_serialization.deserialize(
          metadata,
          item_metadata=item_metadata,
          metrics=metrics,
      )
    elif metadata_type == RootMetadata:
      return root_metadata_serialization.deserialize(metadata)

  def serialize_metadata(
      self,
      metadata: StepMetadata | RootMetadata,
  ) -> checkpoint.SerializedMetadata:
    if isinstance(metadata, StepMetadata):
      return step_metadata_serialization.serialize(metadata)
    elif isinstance(metadata, RootMetadata):
      return root_metadata_serialization.serialize(metadata)

  def get_metadata(
      self,
      metadata_type: type[StepMetadata] | type[RootMetadata],
      custom: dict[str, Any] = None,
  ) -> StepMetadata | RootMetadata:
    """Returns a sample metadata object of `metadata_class`."""
    if custom is None:
      custom = {'a': 1}
    if metadata_type == StepMetadata:
      return StepMetadata(
          format=_SAMPLE_FORMAT,
          item_handlers={'a': 'b'},
          item_metadata=CompositeItemMetadata(a=None),
          metrics={'a': 1},
          performance_metrics=StepStatistics(
              step=None,
              event_type='save',
              reached_preemption=False,
              preemption_received_at=1.0,
          ),
          init_timestamp_nsecs=1,
          commit_timestamp_nsecs=1,
          custom=custom,
      )
    elif metadata_type == RootMetadata:
      return RootMetadata(
          format=_SAMPLE_FORMAT,
          custom=custom,
      )

  def get_metadata_filename(
      self, metadata_type: type[StepMetadata] | type[RootMetadata]
  ) -> str:
    if metadata_type == StepMetadata:
      return checkpoint._STEP_METADATA_FILENAME
    elif metadata_type == RootMetadata:
      return checkpoint._ROOT_METADATA_FILENAME

  def get_metadata_file_path(
      self,
      metadata_type: type[StepMetadata] | type[RootMetadata],
      path: epath.PathLike = None,
  ) -> epath.Path:
    if metadata_type == StepMetadata:
      return checkpoint.step_metadata_file_path(path or self.directory)
    elif metadata_type == RootMetadata:
      return checkpoint.root_metadata_file_path(path or self.directory)

  @parameterized.parameters(True, False)
  def test_read_unknown_path(self, blocking_write: bool):
    self.assertIsNone(
        self.write_metadata_store(blocking_write).read(
            file_path='unknown_checkpoint_path'
        )
    )

  @parameterized.parameters(
      checkpoint.step_metadata_file_path,
      checkpoint.root_metadata_file_path,
  )
  def test_metadata_path_does_not_exist(
      self, file_path_fn: Callable[[epath.PathLike], epath.Path]
  ):
    with self.assertRaisesRegex(FileNotFoundError, 'Path does not exist'):
      file_path_fn('non_existent_metadata_path')

  def test_legacy_root_metadata_file_path(self):
    self.assertEqual(
        checkpoint.root_metadata_file_path(self.directory, legacy=True),
        self.directory / checkpoint._LEGACY_ROOT_METADATA_FILENAME,
    )

  @parameterized.parameters(
      itertools.product(
          [True, False],
          [StepMetadata, RootMetadata],
      )
  )
  def test_write_unknown_file_path(
      self,
      blocking_write: bool,
      metadata_class: type[StepMetadata] | type[RootMetadata],
  ):
    metadata = metadata_class()

    if blocking_write:
      with self.assertRaisesRegex(ValueError, 'Metadata path does not exist'):
        self.write_metadata_store(blocking_write).write(
            file_path='unknown_metadata_path',
            metadata=self.serialize_metadata(metadata),
        )
    else:
      self.write_metadata_store(blocking_write).write(
          file_path='unknown_metadata_path',
          metadata=self.serialize_metadata(metadata),
      )
      try:
        self.write_metadata_store(blocking_write).wait_until_finished()
      except ValueError:
        # We don't want to fail the test because above write's future.result()
        # will raise 'ValueError: Metadata file does not exist ...'.
        pass
      self.assertIsNone(
          self.read_metadata_store(blocking_write).read('unknown_metadata_path')
      )

  @parameterized.parameters(
      itertools.product(
          [True, False],
          [StepMetadata, RootMetadata],
      )
  )
  def test_read_default_values(
      self,
      blocking_write: bool,
      metadata_class: type[StepMetadata] | type[RootMetadata],
  ):
    metadata = metadata_class()

    self.write_metadata_store(blocking_write).write(
        file_path=self.get_metadata_file_path(metadata_class),
        metadata=self.serialize_metadata(metadata),
    )

    self.write_metadata_store(blocking_write).wait_until_finished()

    serialized_metadata = self.write_metadata_store(blocking_write).read(
        file_path=self.get_metadata_file_path(metadata_class),
    )
    self.assertEqual(
        self.deserialize_metadata(metadata_class, serialized_metadata),
        metadata,
    )

  @parameterized.parameters(
      itertools.product(
          [True, False],
          [StepMetadata, RootMetadata],
      )
  )
  def test_read_with_values(
      self,
      blocking_write: bool,
      metadata_class: type[StepMetadata] | type[RootMetadata],
  ):
    metadata = self.get_metadata(metadata_class)

    self.write_metadata_store(blocking_write).write(
        file_path=self.get_metadata_file_path(metadata_class),
        metadata=self.serialize_metadata(metadata),
    )

    self.write_metadata_store(blocking_write).wait_until_finished()

    serialized_metadata = self.write_metadata_store(blocking_write).read(
        file_path=self.get_metadata_file_path(metadata_class),
    )
    self.assertEqual(
        self.deserialize_metadata(metadata_class, serialized_metadata),
        metadata,
    )

  @parameterized.parameters(
      itertools.product(
          [True, False],
          [StepMetadata, RootMetadata],
      )
  )
  def test_read_corrupt_json_data(
      self,
      blocking_write: bool,
      metadata_class: type[StepMetadata] | type[RootMetadata],
  ):
    metadata_file = self.get_metadata_file_path(metadata_class)
    metadata_file.touch()

    self.assertIsNone(
        self.write_metadata_store(blocking_write).read(
            file_path=self.get_metadata_file_path(metadata_class)
        )
    )

  @parameterized.parameters(
      itertools.product(
          [True, False],
          [StepMetadata, RootMetadata],
      )
  )
  def test_update_without_prior_data(
      self,
      blocking_write: bool,
      metadata_class: type[StepMetadata] | type[RootMetadata],
  ):
    self.write_metadata_store(blocking_write).update(
        file_path=self.get_metadata_file_path(metadata_class),
        format=_SAMPLE_FORMAT,
        custom={'a': 1},
    )

    self.write_metadata_store(blocking_write).wait_until_finished()

    serialized_metadata = self.write_metadata_store(blocking_write).read(
        file_path=self.get_metadata_file_path(metadata_class),
    )
    self.assertEqual(
        self.deserialize_metadata(metadata_class, serialized_metadata),
        metadata_class(
            format=_SAMPLE_FORMAT,
            custom={'a': 1},
        ),
    )

  @parameterized.parameters(
      itertools.product(
          [True, False],
          [StepMetadata, RootMetadata],
      )
  )
  def test_update_with_prior_data(
      self,
      blocking_write: bool,
      metadata_class: type[StepMetadata] | type[RootMetadata],
  ):
    metadata = metadata_class(format=_SAMPLE_FORMAT)
    self.write_metadata_store(blocking_write).write(
        file_path=self.get_metadata_file_path(metadata_class),
        metadata=self.serialize_metadata(metadata),
    )

    self.write_metadata_store(blocking_write).update(
        file_path=self.get_metadata_file_path(metadata_class),
        custom={'a': 1},
    )

    self.write_metadata_store(blocking_write).wait_until_finished()

    serialized_metadata = self.write_metadata_store(blocking_write).read(
        file_path=self.get_metadata_file_path(metadata_class)
    )
    self.assertEqual(
        self.deserialize_metadata(metadata_class, serialized_metadata),
        metadata_class(
            format=_SAMPLE_FORMAT,
            custom={'a': 1},
        ),
    )

  @parameterized.parameters(
      itertools.product(
          [True, False],
          [StepMetadata, RootMetadata],
      )
  )
  def test_update_with_unknown_kwargs(
      self,
      blocking_write: bool,
      metadata_class: type[StepMetadata] | type[RootMetadata]
  ):
    metadata = metadata_class()
    self.write_metadata_store(blocking_write).write(
        file_path=self.get_metadata_file_path(metadata_class),
        metadata=self.serialize_metadata(metadata),
    )
    self.write_metadata_store(blocking_write).update(
        file_path=self.get_metadata_file_path(metadata_class),
        format=_SAMPLE_FORMAT,
        blah=2,
    )

    self.write_metadata_store(blocking_write).wait_until_finished()

    serialized_metadata = self.write_metadata_store(blocking_write).read(
        file_path=self.get_metadata_file_path(metadata_class)
    )
    self.assertEqual(
        self.deserialize_metadata(metadata_class, serialized_metadata),
        metadata_class(
            format=_SAMPLE_FORMAT,
            custom={'blah': 2},
        ),
    )

  @parameterized.parameters(
      itertools.product(
          [True, False],
          [StepMetadata, RootMetadata],
      )
  )
  def test_write_with_read_only_store_is_no_op(
      self,
      blocking_write: bool,
      metadata_class: type[StepMetadata] | type[RootMetadata],
  ):
    metadata = metadata_class()

    self.read_metadata_store(blocking_write).write(
        file_path=self.get_metadata_file_path(metadata_class),
        metadata=self.serialize_metadata(metadata),
    )

    serialized_metadata = self.read_metadata_store(blocking_write).read(
        file_path=self.get_metadata_file_path(metadata_class)
    )
    self.assertIsNone(serialized_metadata)

    serialized_metadata = self.write_metadata_store(blocking_write).read(
        file_path=self.get_metadata_file_path(metadata_class)
    )
    self.assertIsNone(serialized_metadata)

  @parameterized.parameters(StepMetadata, RootMetadata)
  def test_non_blocking_write_request_enables_writes(
      self, metadata_class: type[StepMetadata] | type[RootMetadata],
  ):
    metadata = self.get_metadata(metadata_class)

    # setup some data with blocking store.
    self.write_metadata_store(blocking_write=True).write(
        file_path=self.get_metadata_file_path(metadata_class),
        metadata=self.serialize_metadata(metadata),
    )

    serialized_metadata = self.read_metadata_store(blocking_write=True).read(
        file_path=self.get_metadata_file_path(metadata_class)
    )
    self.assertEqual(
        self.deserialize_metadata(metadata_class, serialized_metadata),
        self.get_metadata(metadata_class),
    )

    # write validations
    serialized_metadata = self.serialize_metadata(
        self.get_metadata(metadata_class, custom={'a': 2})
    )
    self.write_metadata_store(blocking_write=False).write(
        file_path=self.get_metadata_file_path(metadata_class),
        metadata=serialized_metadata,
    )

    self.write_metadata_store(blocking_write=False).wait_until_finished()

    serialized_metadata = self.read_metadata_store(blocking_write=False).read(
        file_path=self.get_metadata_file_path(metadata_class)
    )
    self.assertEqual(
        self.deserialize_metadata(metadata_class, serialized_metadata),
        self.get_metadata(metadata_class, custom={'a': 2}),
    )
    serialized_metadata = self.write_metadata_store(blocking_write=False).read(
        file_path=self.get_metadata_file_path(metadata_class)
    )
    self.assertEqual(
        self.deserialize_metadata(metadata_class, serialized_metadata),
        self.get_metadata(metadata_class, custom={'a': 2}),
    )

    # update validations
    self.write_metadata_store(blocking_write=False).update(
        file_path=self.get_metadata_file_path(metadata_class),
        custom={'a': 3},
    )
    self.write_metadata_store(blocking_write=False).wait_until_finished()
    serialized_metadata = self.read_metadata_store(blocking_write=False).read(
        file_path=self.get_metadata_file_path(metadata_class)
    )
    self.assertEqual(
        self.deserialize_metadata(metadata_class, serialized_metadata),
        self.get_metadata(metadata_class, custom={'a': 3}),
    )
    serialized_metadata = self.write_metadata_store(blocking_write=False).read(
        file_path=self.get_metadata_file_path(metadata_class)
    )
    self.assertEqual(
        self.deserialize_metadata(metadata_class, serialized_metadata),
        self.get_metadata(metadata_class, custom={'a': 3}),
    )

  @parameterized.parameters(StepMetadata, RootMetadata)
  def test_metadata_file_path(
      self, metadata_class: type[StepMetadata] | type[RootMetadata]
  ):
    self.assertEqual(
        self.get_metadata_file_path(metadata_class),
        self.directory / self.get_metadata_filename(metadata_class),
    )

    with self.assertRaisesRegex(IOError, 'Path does not exist'):
      self.get_metadata_file_path(metadata_class, path=self.directory / '0')

    metadata_file = self.get_metadata_file_path(metadata_class)
    self.write_metadata_store(blocking_write=True).write(
        file_path=metadata_file,
        metadata=self.serialize_metadata(self.get_metadata(metadata_class)),
    )
    with self.assertRaisesRegex(IOError, 'Path is not a directory'):
      self.get_metadata_file_path(metadata_class, path=metadata_file)

  @parameterized.parameters(True, False)
  def test_pickle(self, blocking_write: bool):
    with self.assertRaisesRegex(TypeError, 'cannot pickle'):
      _ = pickle.dumps(self.write_metadata_store(blocking_write))
    _ = pickle.dumps(self.read_metadata_store(blocking_write))

  @parameterized.parameters(
      ({'format': int()},),
      ({'custom': list()},),
      ({'custom': {int(): None}},),
  )
  def test_deserialize_wrong_types_root_metadata(
      self, wrong_metadata: checkpoint.SerializedMetadata
  ):
    with self.assertRaises(ValueError):
      self.deserialize_metadata(RootMetadata, wrong_metadata)

  @parameterized.parameters(
      ({'format': int()},),
      ({'item_handlers': list()},),
      ({'item_handlers': {int(): None}},),
      ({'item_metadata': list()},),
      ({'item_metadata': {int(): None}},),
      ({'metrics': list()},),
      ({'metrics': {int(): None}},),
      ({'performance_metrics': list()},),
      ({'performance_metrics': {int(): float()}},),
      ({'performance_metrics': {str(): int()}},),
      ({'init_timestamp_nsecs': float()},),
      ({'commit_timestamp_nsecs': float()},),
      ({'custom': list()},),
      ({'custom': {int(): None}},),
  )
  def test_deserialize_wrong_types_step_metadata(
      self, wrong_metadata: checkpoint.SerializedMetadata
  ):
    with self.assertRaises(ValueError):
      self.deserialize_metadata(StepMetadata, wrong_metadata)

  @parameterized.parameters(
      (
          RootMetadata(custom={'a': None}),
          {'custom': {'a': None}}
      ),
      (
          RootMetadata(format=_SAMPLE_FORMAT),
          {'format': _SAMPLE_FORMAT}
      ),
      (
          StepMetadata(format=_SAMPLE_FORMAT),
          {'format': _SAMPLE_FORMAT},
      ),
      (
          StepMetadata(item_handlers={'a': 'a_handler'}),
          {'item_handlers': {'a': 'a_handler'}},
      ),
      (
          StepMetadata(custom={'blah': 123}),
          {'custom': {'blah': 123}},
      ),
  )
  def test_only_serialize_non_default_metadata_values(
      self,
      metadata: StepMetadata | RootMetadata,
      expected_serialized_metadata: dict[str, Any],
  ):
    self.assertEqual(
        self.serialize_metadata(metadata), expected_serialized_metadata
    )

  @parameterized.parameters(StepMetadata, RootMetadata)
  def test_unknown_key_in_metadata(
      self, metadata_class: type[StepMetadata] | type[RootMetadata],
  ):
    metadata = metadata_class(
        format=_SAMPLE_FORMAT,
        custom={'a': 1},
    )
    serialized_metadata = self.serialize_metadata(metadata)
    serialized_metadata['blah'] = 2
    with self.assertRaisesRegex(
        ValueError, 'Provided metadata contains unknown key blah'
    ):
      self.deserialize_metadata(metadata_class, serialized_metadata)

  @parameterized.named_parameters(
      ('single', 'a_handler'),
      ('composite', {'a': 'a_handler'}),
  )
  def test_deserialize_item_handlers(self, item_handlers):
    serialized_metadata = {'item_handlers': item_handlers}
    metadata = self.deserialize_metadata(StepMetadata, serialized_metadata)
    self.assertEqual(metadata.item_handlers, item_handlers)

  @parameterized.named_parameters(
      ('single', 'a_metadata'),
      ('composite', {'a': 'a_metadata'}),
  )
  def test_deserialize_item_metadata(self, item_metadata):
    serialized_metadata = {'item_metadata': item_metadata}
    metadata = self.deserialize_metadata(StepMetadata, serialized_metadata)
    self.assertEqual(metadata.item_metadata, item_metadata)

  @parameterized.named_parameters(
      ('single_same_kwarg', 'a', 'a', 'a'),
      ('single_different_kwarg', 'a', 'b', 'b'),
      ('composite_same_kwarg', {'A': 'a'}, {'A': 'a'}, {'A': 'a'}),
      ('composite_different_kwarg', {'A': 'a'}, {'A': 'b'}, {'A': 'b'}),
      ('composite_new_kwarg', {'A': 'a'}, {'B': 'b'}, {'A': 'a', 'B': 'b'}),
  )
  def test_deserialize_item_metadata_with_item_metadata_kwarg(
      self, item_metadata, item_metadata_kwarg, expected_item_metadata
  ):
    serialized_metadata = {'item_metadata': item_metadata}
    metadata = self.deserialize_metadata(
        StepMetadata,
        serialized_metadata,
        item_metadata=item_metadata_kwarg,
    )
    self.assertEqual(metadata.item_metadata, expected_item_metadata)

  @parameterized.named_parameters(
      ('single', 'a_metadata', {'a': 'a_metadata'}),
      ('composite', {'a': 'a_metadata'}, 'a_metadata'),
  )
  def test_deserialize_item_metadata_with_item_metadata_kwarg_mismatch(
      self, item_metadata, item_metadata_kwarg
  ):
    serialized_metadata = {'item_metadata': item_metadata}
    with self.assertRaisesRegex(
        ValueError,
        'Provided item_metadata is of type {}, but the serialized '
        'StepMetadata.item_metadata is of type {}. Cannot deserialize '
        'mismatched types.'.format(
            type(item_metadata_kwarg), type(item_metadata)
        ),
    ):
      self.deserialize_metadata(
          StepMetadata,
          serialized_metadata,
          item_metadata=item_metadata_kwarg,
      )

  @parameterized.parameters(
      ('metrics', 1, dict, int),
  )
  def test_validate_type_wrong_type(
      self, kwarg_name, kwarg_value, expected_type, wrong_type
  ):
    with self.assertRaisesRegex(
        ValueError,
        f'Object must be of type {expected_type}, got {wrong_type}.',
    ):
      self.deserialize_metadata(StepMetadata, {}, **{kwarg_name: kwarg_value})

  @parameterized.parameters(
      ({'format': 1}, 'format', str, int),
      ({'metrics': 1}, 'metrics', dict, int),
      ({'performance_metrics': 1}, 'performance_metrics', dict, int),
      ({'init_timestamp_nsecs': 'a'}, 'init_timestamp_nsecs', int, str),
      ({'commit_timestamp_nsecs': 'a'}, 'commit_timestamp_nsecs', int, str),
      ({'custom': 1}, 'custom', dict, int),
  )
  def test_validate_field_wrong_type(
      self, step_metadata, field_name, expected_field_type, wrong_field_type
  ):
    with self.assertRaisesRegex(
        ValueError,
        f'Metadata field "{field_name}" must be of type '
        f'{expected_field_type}, got {wrong_field_type}.',
    ):
      self.deserialize_metadata(StepMetadata, step_metadata)

  @parameterized.parameters(
      ({'item_handlers': 1}, 'item_handlers', int),
      ({'item_metadata': 1}, 'item_metadata', int),
  )
  def test_validate_field_sequence_wrong_type(
      self, step_metadata, field_name, wrong_field_type
  ):
    with self.assertRaisesRegex(
        ValueError,
        f'Metadata field "{field_name}" must be any one of types '
        r'\[.+]'
        f', got {wrong_field_type}.',
    ):
      self.deserialize_metadata(StepMetadata, step_metadata)

  @parameterized.parameters(
      ({'item_handlers': {1: 'a'}}, 'item_handlers', str, int),
      ({'item_metadata': {1: 'a'}}, 'item_metadata', str, int),
      ({'metrics': {1: 'a'}}, 'metrics', str, int),
      (
          {'performance_metrics': {1: 1.0}},
          'performance_metrics',
          str,
          int,
      ),
      ({'custom': {1: 'a'}}, 'custom', str, int),
  )
  def test_validate_dict_entry_wrong_key_type(
      self, step_metadata, field_name, expected_key_type, wrong_key_type
  ):
    with self.assertRaisesRegex(
        ValueError,
        f'Metadata field "{field_name}" keys must be of type '
        f'{expected_key_type}, got {wrong_key_type}.',
    ):
      self.deserialize_metadata(StepMetadata, step_metadata)

  @parameterized.parameters(
      ({'performance_metrics': {'a': 1}}, 'performance_metrics', float, int),
  )
  def test_validate_dict_entry_wrong_value_type(
      self, step_metadata, field_name, expected_value_type, wrong_value_type
  ):
    with self.assertRaisesRegex(
        ValueError,
        f'Metadata field "{field_name}" values must be of type '
        f'{expected_value_type}, got {wrong_value_type}.',
    ):
      self.deserialize_metadata(StepMetadata, step_metadata)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

from absl.testing import absltest
from absl.testing import parameterized
import chex
import jax
from orbax.checkpoint._src.metadata import tree as tree_metadata
from orbax.checkpoint._src.serialization import type_handlers
from orbax.checkpoint._src.serialization import types
from orbax.checkpoint._src.testing import test_tree_utils
from orbax.checkpoint._src.tree import utils as tree_utils


def _to_param_infos(
    tree: Any,
    pytree_metadata_options: tree_metadata.PyTreeMetadataOptions,
):
  return jax.tree.map(
      # Other properties are not relevant.
      lambda x: types.ParamInfo(
          value_typestr=types.get_param_typestr(
              x,
              type_handlers.GLOBAL_TYPE_HANDLER_REGISTRY,
              pytree_metadata_options,
          )
      ),
      tree,
      is_leaf=tree_utils.is_empty_or_leaf,
  )


class InternalTreeMetadataEntryTest(parameterized.TestCase):

  @parameterized.product(
      test_pytree=test_tree_utils.TEST_PYTREES,
      pytree_metadata_options=[
          tree_metadata.PyTreeMetadataOptions(support_rich_types=False),
          tree_metadata.PyTreeMetadataOptions(support_rich_types=True),
      ],
  )
  def test_as_nested_tree(
      self,
      test_pytree: test_tree_utils.TestPyTree,
      pytree_metadata_options: tree_metadata.PyTreeMetadataOptions,
  ):
    tree = test_pytree.provide_tree()
    original_internal_tree_metadata = tree_metadata.InternalTreeMetadata.build(
        param_infos=_to_param_infos(tree, pytree_metadata_options),
        pytree_metadata_options=pytree_metadata_options,
    )
    json_object = original_internal_tree_metadata.to_json()
    restored_internal_tree_metadata = (
        tree_metadata.InternalTreeMetadata.from_json(
            json_object, pytree_metadata_options
        )
    )

    if pytree_metadata_options.support_rich_types:
      expected_tree_metadata = (
          test_pytree.expected_nested_tree_metadata_with_rich_types
      )
    else:
      expected_tree_metadata = test_pytree.expected_nested_tree_metadata
    restored_tree_metadata = restored_internal_tree_metadata.as_nested_tree()
    chex.assert_trees_all_equal(restored_tree_metadata, expected_tree_metadata)

  @parameterized.product(
      test_pytree=test_tree_utils.TEST_PYTREES,
      pytree_metadata_options_switch=[
          (
              tree_metadata.PyTreeMetadataOptions(support_rich_types=False),
              tree_metadata.PyTreeMetadataOptions(support_rich_types=True),
          ),
          (
              tree_metadata.PyTreeMetadataOptions(support_rich_types=True),
              tree_metadata.PyTreeMetadataOptions(support_rich_types=False),
          ),
      ],
  )
  def test_switching_between_support_rich_types(
      self,
      test_pytree: test_tree_utils.TestPyTree,
      pytree_metadata_options_switch: tuple[
          tree_metadata.PyTreeMetadataOptions,
          tree_metadata.PyTreeMetadataOptions,
      ],
  ):
    write_pytree_metadata_options, read_pytree_metadata_options = (
        pytree_metadata_options_switch
    )
    if (
        write_pytree_metadata_options.support_rich_types
        == read_pytree_metadata_options.support_rich_types
    ):
      self.fail(
          'This test is only meant for scenarios when support_rich_types will'
          ' be different between write and read.'
      )
    else:
      expected_tree_metadata = test_pytree.expected_nested_tree_metadata
    tree = test_pytree.provide_tree()

    original_internal_tree_metadata = tree_metadata.InternalTreeMetadata.build(
        param_infos=_to_param_infos(tree, write_pytree_metadata_options),
        pytree_metadata_options=write_pytree_metadata_options,
    )
    json_object = original_internal_tree_metadata.to_json()
    restored_internal_tree_metadata = (
        tree_metadata.InternalTreeMetadata.from_json(
            json_object, read_pytree_metadata_options
        )
    )

    restored_tree_metadata = restored_internal_tree_metadata.as_nested_tree()
    chex.assert_trees_all_equal(restored_tree_metadata, expected_tree_metadata)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A global registry of checkpoint handler types."""

from typing import Type

from absl import logging
from orbax.checkpoint._src.handlers import checkpoint_handler


CheckpointHandler = checkpoint_handler.CheckpointHandler


class HandlerTypeRegistry:
  """A registry mapping handler type strings to handler types."""

  def __init__(self):
    self._registry = {}

  def add(
      self,
      handler_typestr: str,
      handler_type: Type[CheckpointHandler],
  ) -> None:
    """Adds an entry to the registry."""
    if handler_typestr in self._registry:
      if self._registry[handler_typestr] != handler_type:
        raise ValueError(
            f'Handler type string "{handler_typestr}" already exists in the '
            f'registry with type {self._registry[handler_typestr]}. '
            f'Cannot add type {handler_type}.'
        )
      else:
        logging.info(
            'Handler type string "%s" already exists in the registry with '
            'associated type %s.',
            handler_typestr,
            self._registry[handler_typestr],
        )
        return
    self._registry[handler_typestr] = handler_type

  def get(
      self,
      handler_typestr: str,
  ) -> Type[CheckpointHandler]:
    """Gets an entry from the registry."""
    if handler_typestr not in self._registry:
      raise ValueError(
          f'Handler type string "{handler_typestr}" not found in the registry.'
      )
    return self._registry[handler_typestr]


_GLOBAL_HANDLER_TYPE_REGISTRY = HandlerTypeRegistry()


def register_handler_type(handler_cls):
  _GLOBAL_HANDLER_TYPE_REGISTRY.add(handler_cls.typestr(), handler_cls)
  return handler_cls


def get_handler_type(handler_typestr: str) -> Type[CheckpointHandler]:
  return _GLOBAL_HANDLER_TYPE_REGISTRY.get(handler_typestr)



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Internal IO utilities for metadata of a checkpoint at step level."""

import dataclasses
from typing import Any, Mapping

from absl import logging
from orbax.checkpoint._src.logging import step_statistics
from orbax.checkpoint._src.metadata import checkpoint
from orbax.checkpoint._src.metadata import metadata_serialization_utils as utils

SerializedMetadata = checkpoint.SerializedMetadata
StepMetadata = checkpoint.StepMetadata
CompositeCheckpointHandlerTypeStrs = (
    checkpoint.CompositeCheckpointHandlerTypeStrs
)
CheckpointHandlerTypeStr = checkpoint.CheckpointHandlerTypeStr
CompositeItemMetadata = checkpoint.CompositeItemMetadata
SingleItemMetadata = checkpoint.SingleItemMetadata
StepStatistics = step_statistics.SaveStepStatistics


def serialize(metadata: StepMetadata) -> SerializedMetadata:
  """Serializes `metadata` to a dictionary."""

  # Part of the StepMetadata api for user convenience, but not saved to disk.
  if (metadata.item_metadata is not None and
      isinstance(metadata.item_metadata, CompositeItemMetadata)):
    just_item_names = {k: None for k in metadata.item_metadata.keys()}
  else:
    just_item_names = None

  # Save only float performance metrics.
  performance_metrics = metadata.performance_metrics
  float_metrics = {
      metric: val
      for metric, val in dataclasses.asdict(performance_metrics).items()
      if isinstance(val, float)
  }

  serialized_metadata = {}
  if metadata.format is not None:
    serialized_metadata['format'] = metadata.format
  if metadata.item_handlers:
    serialized_metadata['item_handlers'] = metadata.item_handlers
  if just_item_names is not None:
    serialized_metadata['item_metadata'] = just_item_names
  if metadata.metrics:
    serialized_metadata['metrics'] = metadata.metrics
  if float_metrics:
    serialized_metadata['performance_metrics'] = float_metrics
  if metadata.init_timestamp_nsecs is not None:
    serialized_metadata['init_timestamp_nsecs'] = metadata.init_timestamp_nsecs
  if metadata.commit_timestamp_nsecs is not None:
    serialized_metadata['commit_timestamp_nsecs'] = (
        metadata.commit_timestamp_nsecs
    )
  if metadata.custom:
    serialized_metadata['custom'] = metadata.custom

  return serialized_metadata


def deserialize(
    metadata_dict: SerializedMetadata,
    item_metadata: CompositeItemMetadata | SingleItemMetadata | None = None,
    metrics: dict[str, Any] | None = None,
) -> StepMetadata:
  """Deserializes `metadata_dict` and other kwargs to `StepMetadata`."""
  validated_metadata_dict = {}

  utils.validate_field(metadata_dict, 'format', str)
  validated_metadata_dict['format'] = metadata_dict.get('format', None)

  utils.validate_field(metadata_dict, 'item_handlers', [dict, str])
  item_handlers = metadata_dict.get('item_handlers')
  if isinstance(item_handlers, CompositeCheckpointHandlerTypeStrs):
    for k in metadata_dict.get('item_handlers', {}) or {}:
      utils.validate_dict_entry(metadata_dict, 'item_handlers', k, str)
    validated_metadata_dict['item_handlers'] = item_handlers
  elif isinstance(item_handlers, CheckpointHandlerTypeStr):
    validated_metadata_dict['item_handlers'] = item_handlers

  utils.validate_field(metadata_dict, 'item_metadata', [dict, str])
  dict_item_metadata = metadata_dict.get('item_metadata')
  if (item_metadata is not None and dict_item_metadata is not None and
      type(item_metadata) is not type(dict_item_metadata)):
    raise ValueError(
        f'Provided item_metadata is of type {type(item_metadata)}, but the '
        'serialized StepMetadata.item_metadata is of '
        f'type {type(dict_item_metadata)}. Cannot deserialize mismatched types.'
    )
  if isinstance(dict_item_metadata, Mapping):
    for k in dict_item_metadata or {}:
      utils.validate_dict_entry(metadata_dict, 'item_metadata', k, str)
    validated_metadata_dict['item_metadata'] = dict_item_metadata
    if item_metadata is not None:
      if validated_metadata_dict['item_metadata'] is None:
        validated_metadata_dict['item_metadata'] = {}
      for k, v in item_metadata.items():
        utils.validate_dict_entry(metadata_dict, 'item_metadata', k, str)
        if k in validated_metadata_dict['item_metadata']:
          logging.warning(
              'Overwriting item_metadata entry {%s: %s} found in serialized '
              'StepMetadata with the provided item_metadata entry {%s: %s}.',
              k, validated_metadata_dict['item_metadata'][k], k, v
          )
        validated_metadata_dict['item_metadata'][k] = v
  elif item_metadata is not None:
    logging.warning(
        'Overwriting item_metadata found in serialized StepMetadata (%s) '
        'with the provided item_metadata (%s).',
        dict_item_metadata, item_metadata
    )
    validated_metadata_dict['item_metadata'] = item_metadata
  else:
    validated_metadata_dict['item_metadata'] = dict_item_metadata

  utils.validate_field(metadata_dict, 'metrics', dict)
  for k in metadata_dict.get('metrics', {}) or {}:
    utils.validate_dict_entry(metadata_dict, 'metrics', k, str)
  validated_metadata_dict['metrics'] = metadata_dict.get('metrics', {})
  if metrics is not None:
    utils.validate_type(metrics, dict)
    for k, v in metrics.items():
      utils.validate_dict_entry(metadata_dict, 'metrics', k, str)
      validated_metadata_dict['metrics'][k] = v

  utils.validate_field(metadata_dict, 'performance_metrics', dict)
  for k in metadata_dict.get('performance_metrics', {}) or {}:
    utils.validate_dict_entry(
        metadata_dict, 'performance_metrics', k, str, float
    )
  validated_metadata_dict['performance_metrics'] = StepStatistics(
      **metadata_dict.get('performance_metrics', {})
  )

  utils.validate_field(metadata_dict, 'init_timestamp_nsecs', int)
  validated_metadata_dict['init_timestamp_nsecs'] = (
      metadata_dict.get('init_timestamp_nsecs', None)
  )

  utils.validate_field(metadata_dict, 'commit_timestamp_nsecs', int)
  validated_metadata_dict['commit_timestamp_nsecs'] = (
      metadata_dict.get('commit_timestamp_nsecs', None)
  )

  utils.validate_field(metadata_dict, 'custom', dict)
  for k in metadata_dict.get('custom', {}) or {}:
    utils.validate_dict_entry(metadata_dict, 'custom', k, str)
  validated_metadata_dict['custom'] = metadata_dict.get('custom', {})

  for k in metadata_dict:
    if k not in validated_metadata_dict:
      if 'custom' in metadata_dict and metadata_dict['custom']:
        raise ValueError(
            'Provided metadata contains unknown key %s, and the custom field '
            'is already defined.' % k
        )
      logging.warning(
          'Provided metadata contains unknown key %s. Adding it to custom.', k
      )
      validated_metadata_dict['custom'][k] = metadata_dict[k]

  return StepMetadata(**validated_metadata_dict)



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for empty values as leafs in the checkpoint tree."""

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np
import optax
from orbax.checkpoint._src.metadata import empty_values
from orbax.checkpoint._src.metadata import pytree_metadata_options
from orbax.checkpoint._src.testing import test_tree_utils


class EmptyValuesTest(parameterized.TestCase):

  @parameterized.parameters(
      (1, False, False),
      (np.zeros(0), False, False),
      (dict(), True, True),
      ({}, True, True),
      ({"a": {}}, False, False),
      ([], True, True),
      ([[]], False, False),
      (tuple(), True, True),
      ((tuple(),), False, False),
      (None, True, True),
      ((1, 2), False, False),
      (test_tree_utils.EmptyNamedTuple(), False, True),
      (optax.EmptyState(), False, True),
      (test_tree_utils.MuNu(mu=None, nu=None), False, False),
      (test_tree_utils.NamedTupleWithNestedAttributes(), False, False),
      (
          test_tree_utils.NamedTupleWithNestedAttributes(nested_dict={}),
          False,
          False,
      ),
      (test_tree_utils.MyEmptyChex(), True, True),
      (test_tree_utils.MyChex(), False, False),
      (test_tree_utils.MyChex(my_np_array=np.array([])), False, False),
      (test_tree_utils.MyEmptyClass(), False, False),
      (test_tree_utils.MyClass(), False, False),
      (test_tree_utils.MyClass(a=None, b=None), False, False),
      (test_tree_utils.MyClass(a=None, b=np.zeros(1)), False, False),
      (test_tree_utils.MyEmptyFlax(), False, False),  # TODO: b/378905913 - fix
      (test_tree_utils.MyFlax(), False, False),
      (
          test_tree_utils.MyFlax(
              my_jax_array=None, my_nested_mapping=None, my_sequence=None
          ),
          False,
          False,
      ),
      (test_tree_utils.MyFlax(my_nested_mapping={"a": 1}), False, False),
      (test_tree_utils.MyEmptyDataClass(), False, False),
      (test_tree_utils.MyDataClass(), False, False),
      (
          test_tree_utils.MyDataClass(
              my_jax_array=None,
              my_np_array=None,
              my_empty_dataclass=None,
              my_chex=None,
          ),
          False,
          False,
      ),
  )
  def test_is_supported_empty_value(self, value, expected, expected_rich_type):
    with self.subTest("legacy_metadata"):
      self.assertEqual(
          expected,
          empty_values.is_supported_empty_value(
              value,
              pytree_metadata_options.PyTreeMetadataOptions(
                  support_rich_types=False
              ),
          ),
      )
    with self.subTest("rich_typed_metadata"):
      self.assertEqual(
          expected_rich_type,
          empty_values.is_supported_empty_value(
              value,
              pytree_metadata_options.PyTreeMetadataOptions(
                  support_rich_types=True
              ),
          ),
      )

  @parameterized.parameters(
      (1, ValueError(), ValueError()),
      (np.zeros(0), ValueError(), ValueError()),
      (dict(), empty_values.RESTORE_TYPE_DICT, empty_values.RESTORE_TYPE_DICT),
      ({}, empty_values.RESTORE_TYPE_DICT, empty_values.RESTORE_TYPE_DICT),
      ({"a": {}}, ValueError(), ValueError()),
      ([], empty_values.RESTORE_TYPE_LIST, empty_values.RESTORE_TYPE_LIST),
      ([[]], ValueError(), ValueError()),
      (
          tuple(),
          empty_values.RESTORE_TYPE_TUPLE,
          empty_values.RESTORE_TYPE_TUPLE,
      ),
      ((tuple(),), ValueError(), ValueError()),
      (None, empty_values.RESTORE_TYPE_NONE, empty_values.RESTORE_TYPE_NONE),
      ((1, 2), ValueError(), ValueError()),
      (
          test_tree_utils.EmptyNamedTuple(),
          ValueError(),
          empty_values.RESTORE_TYPE_NAMED_TUPLE,
      ),
      (optax.EmptyState(), ValueError(), empty_values.RESTORE_TYPE_NAMED_TUPLE),
      (test_tree_utils.MuNu(mu=None, nu=None), ValueError(), ValueError()),
      (
          test_tree_utils.NamedTupleWithNestedAttributes(),
          ValueError(),
          ValueError(),
      ),
      (
          test_tree_utils.NamedTupleWithNestedAttributes(nested_dict={}),
          ValueError(),
          ValueError(),
      ),
      (
          test_tree_utils.MyEmptyChex(),
          empty_values.RESTORE_TYPE_DICT,
          empty_values.RESTORE_TYPE_DICT,
      ),
      (test_tree_utils.MyChex(), ValueError(), ValueError()),
      (
          test_tree_utils.MyChex(my_np_array=np.array([])),
          ValueError(),
          ValueError(),
      ),
      (test_tree_utils.MyEmptyClass(), ValueError(), ValueError()),
      (test_tree_utils.MyClass(), ValueError(), ValueError()),
      (test_tree_utils.MyClass(a=None, b=None), ValueError(), ValueError()),
      (
          test_tree_utils.MyClass(a=None, b=np.zeros(1)),
          ValueError(),
          ValueError(),
      ),
      (
          test_tree_utils.MyEmptyFlax(),  # TODO: b/378905913 - fix
          ValueError(),
          ValueError(),
      ),
      (test_tree_utils.MyFlax(), ValueError(), ValueError()),
      (
          test_tree_utils.MyFlax(
              my_jax_array=None, my_nested_mapping=None, my_sequence=None
          ),
          ValueError(),
          ValueError(),
      ),
      (
          test_tree_utils.MyFlax(my_nested_mapping={"a": 1}),
          ValueError(),
          ValueError(),
      ),
      (test_tree_utils.MyEmptyDataClass(), ValueError(), ValueError()),
      (test_tree_utils.MyDataClass(), ValueError(), ValueError()),
      (
          test_tree_utils.MyDataClass(
              my_jax_array=None,
              my_np_array=None,
              my_empty_dataclass=None,
              my_chex=None,
          ),
          ValueError(),
          ValueError(),
      ),
  )
  def test_get_empty_value_typestr(self, value, expected, expected_rich_type):
    with self.subTest("legacy_metadata"):
      if isinstance(expected, ValueError):
        with self.assertRaises(ValueError):
          empty_values.get_empty_value_typestr(
              value,
              pytree_metadata_options.PyTreeMetadataOptions(
                  support_rich_types=False
              ),
          )
      else:
        self.assertEqual(
            expected,
            empty_values.get_empty_value_typestr(
                value,
                pytree_metadata_options.PyTreeMetadataOptions(
                    support_rich_types=False
                ),
            ),
        )
    with self.subTest("rich_typed_metadata"):
      if isinstance(expected_rich_type, ValueError):
        with self.assertRaises(ValueError):
          empty_values.get_empty_value_typestr(
              value,
              pytree_metadata_options.PyTreeMetadataOptions(
                  support_rich_types=True
              ),
          )
      else:
        self.assertEqual(
            expected_rich_type,
            empty_values.get_empty_value_typestr(
                value,
                pytree_metadata_options.PyTreeMetadataOptions(
                    support_rich_types=True
                ),
            ),
        )

  @parameterized.parameters(
      (empty_values.RESTORE_TYPE_NONE, None, None),
      (empty_values.RESTORE_TYPE_DICT, {}, {}),
      (empty_values.RESTORE_TYPE_LIST, [], []),
      (empty_values.RESTORE_TYPE_TUPLE, tuple(), tuple()),
      (
          empty_values.RESTORE_TYPE_NAMED_TUPLE,
          None,
          empty_values.OrbaxEmptyNamedTuple(),
      ),
      (empty_values.RESTORE_TYPE_UNKNOWN, ValueError(), ValueError()),
      ("blah", ValueError(), ValueError()),
  )
  def test_get_empty_value_from_typestr(
      self, typestr, expected, expected_rich_type
  ):
    with self.subTest("legacy_metadata"):
      if isinstance(expected, ValueError):
        with self.assertRaises(ValueError):
          empty_values.get_empty_value_from_typestr(
              typestr,
              pytree_metadata_options.PyTreeMetadataOptions(
                  support_rich_types=False
              ),
          )
      else:
        self.assertEqual(
            expected,
            empty_values.get_empty_value_from_typestr(
                typestr,
                pytree_metadata_options.PyTreeMetadataOptions(
                    support_rich_types=False
                ),
            ),
        )
    with self.subTest("rich_typed_metadata"):
      if isinstance(expected_rich_type, ValueError):
        with self.assertRaises(ValueError):
          empty_values.get_empty_value_from_typestr(
              typestr,
              pytree_metadata_options.PyTreeMetadataOptions(
                  support_rich_types=True
              ),
          )
      else:
        self.assertEqual(
            expected_rich_type,
            empty_values.get_empty_value_from_typestr(
                typestr,
                pytree_metadata_options.PyTreeMetadataOptions(
                    support_rich_types=True
                ),
            ),
        )


if __name__ == "__main__":
  absltest.main()
